#include "prechdr.hpp"
#include "maindlg.hpp"


static const int   knPenWidth         = 2;
static const UINT  WMU_SOCKET_CONNECT = WM_USER + 101;
static const UINT  WMU_SOCKET_CLOSED  = WM_USER + 102;
static const UINT  WMU_SOCKET_ERROR   = WM_USER + 103;


TMainDialog::TMainDialog(CWnd* pParent)
    : CDialog(IDD_MAINWINDOW, pParent)
{
    m_bConnected = FALSE;
    m_bCaptured  = FALSE;
    m_DrawColor  = RGB(0, 0, 0);  // black

    m_CommunicationSocket.SetNotification(this,
                                          WMU_SOCKET_CONNECT,
                                          WMU_SOCKET_CLOSED,
                                          WMU_SOCKET_ERROR);
}  // TMainDialog::TMainDialog


TMainDialog::~TMainDialog()
{
}  // TMainDialog::~TMainDialog


void TMainDialog::UpdateControls()
{
    GetDlgItem(IDC_BUTTON_CONNECT)->EnableWindow(!m_bConnected);
    GetDlgItem(IDC_BUTTON_LISTEN)->EnableWindow(!m_bConnected);
    GetDlgItem(IDC_BUTTON_CLOSE)->EnableWindow(m_bConnected);
}  // TMainDialog::UpdateControls


typedef struct tagREADSOCKETDATA_ASYNC
{
    HEADER_DATA  local_toRead;
} READSOCKETDATA_ASYNC, * LPREADSOCKETDATA_ASYNC;


void TMainDialog::ReadSocketData()
{
    ASSERT(m_bConnected);

    LPREADSOCKETDATA_ASYNC  pReadSocketData;
    
    TRY
    {
        pReadSocketData = new READSOCKETDATA_ASYNC;
        m_CommunicationSocket.Read(&pReadSocketData->local_toRead,
                                   sizeof(pReadSocketData->local_toRead),
                                   ReadSocketData_Completion1,
                                   this,
                                   pReadSocketData);
    }
    CATCH_ALL(e)
    {
        MessageBox("Out of memory");
        OnClicked_Close();
    }
    END_CATCH_ALL        
}  // TMainDialog::ReadSocketData


void TMainDialog::ReadSocketData_Completion1(void* pTMainDialog, void* pStruct, int nResult)
{
    TMainDialog*            pThis           = (TMainDialog*)pTMainDialog;
    LPREADSOCKETDATA_ASYNC  pReadSocketData = (LPREADSOCKETDATA_ASYNC)pStruct;

    ASSERT(pThis != NULL);
    ASSERT(pReadSocketData != NULL);

    if(nResult == knResultOK)
    {
	    if(pReadSocketData->local_toRead.message != kchDoneMessage)
	    {
	        pThis->ProcessMessage(pReadSocketData->local_toRead.message,
	                              ReadSocketData_Completion2,
	                              pThis,
	                              pReadSocketData);
	    }
	    else
	    {
	        delete pReadSocketData;
	        pReadSocketData = NULL;
	
	        // normal close
	        pThis->OnClicked_Close();
	        pThis->MessageBox("Finished watching other side.");
	    }
    }
    else
    {
        delete pReadSocketData;
        pReadSocketData = NULL;
    }
}  // TMainDialog::ReadSocketData_Completion1


void TMainDialog::ReadSocketData_Completion2(void* pTMainDialog, void* pStruct, int nResult)
{
    TMainDialog*            pThis           = (TMainDialog*)pTMainDialog;
    LPREADSOCKETDATA_ASYNC  pReadSocketData = (LPREADSOCKETDATA_ASYNC)pStruct;

    ASSERT(pThis != NULL);
    ASSERT(pReadSocketData != NULL);

    if(nResult == knResultOK)
    {
	    pThis->m_CommunicationSocket.Read(&pReadSocketData->local_toRead,
	                                      sizeof(pReadSocketData->local_toRead),
	                                      ReadSocketData_Completion1,
	                                      pThis,
	                                      pReadSocketData);
    }
    else
    {
        delete pReadSocketData;
        pReadSocketData = NULL;
    }
}  // TMainDialog::ReadSocketData_Completion2


typedef struct tagPROCESSMESSAGE_ASYNC
{
    CompletionFn  param_pfnCompletion;
    void*         param_pRefCon1;
    void*         param_pRefCon2;
} PROCESSMESSAGE_ASYNC, * LPPROCESSMESSAGE_ASYNC;


void TMainDialog::ProcessMessage(char chMessage, CompletionFn pfnCompletion, void* pRefCon1, void* pRefCon2)
{
    ASSERT(pfnCompletion != NULL);
    
    LPPROCESSMESSAGE_ASYNC  pProcessMessage;
    
    TRY
    {
        pProcessMessage = new PROCESSMESSAGE_ASYNC;
        pProcessMessage->param_pfnCompletion = pfnCompletion;
        pProcessMessage->param_pRefCon1      = pRefCon1;
        pProcessMessage->param_pRefCon2      = pRefCon2;

        switch(chMessage)
        {
            case(kchPixelMessage):
                ProcessPixelMessage(ProcessMessage_Completion,
                                    this,
                                    pProcessMessage);
                break;

            case(kchLineMessage):
                ProcessLineMessage(ProcessMessage_Completion,
                                   this,
                                   pProcessMessage);
                break;
        }
    }
    CATCH_ALL(e)
    {
        // let the callback do its cleanup
        (*pfnCompletion)(pRefCon1, pRefCon2, knResultAborted);
        MessageBox("Out of memory");
        OnClicked_Close();
    }
    END_CATCH_ALL
}  // TMainDialog::ProcessMessage


void TMainDialog::ProcessMessage_Completion(void* pTMainDialog, void* pStruct, int nResult)
{
    TMainDialog*            pThis        = (TMainDialog*)pTMainDialog;
    LPPROCESSMESSAGE_ASYNC  pProcessData = (LPPROCESSMESSAGE_ASYNC)pStruct;

    ASSERT(pThis != NULL);
    ASSERT(pProcessData != NULL);
    ASSERT(pProcessData->param_pfnCompletion != NULL);

    (*pProcessData->param_pfnCompletion)(pProcessData->param_pRefCon1,
                                         pProcessData->param_pRefCon2,
                                         nResult);

    delete pProcessData;
    pProcessData = NULL;
}  // TMainDialog::ProcessMessage_Completion


typedef struct tagPROCESSPIXELMESSAGE_ASYNC
{
    CompletionFn  param_pfnCompletion;
    void*         param_pRefCon1;
    void*         param_pRefCon2;
    PIXEL_DATA    local_toRead;
} PROCESSPIXELMESSAGE_ASYNC, * LPPROCESSPIXELMESSAGE_ASYNC;


void TMainDialog::ProcessPixelMessage(CompletionFn pfnCompletion, void* pRefCon1, void* pRefCon2)
{
    ASSERT(pfnCompletion != NULL);

    LPPROCESSPIXELMESSAGE_ASYNC  pProcessPixelMessage;
    
    TRY
    {
        pProcessPixelMessage = new PROCESSPIXELMESSAGE_ASYNC;
        pProcessPixelMessage->param_pfnCompletion = pfnCompletion;
        pProcessPixelMessage->param_pRefCon1      = pRefCon1;
        pProcessPixelMessage->param_pRefCon2      = pRefCon2;

        m_CommunicationSocket.Read(&pProcessPixelMessage->local_toRead,
                                   sizeof(pProcessPixelMessage->local_toRead),
                                   ProcessPixelMessage_Completion,
                                   this,
                                   pProcessPixelMessage);
    }
    CATCH_ALL(e)
    {
        // let the callback do its cleanup
        (*pfnCompletion)(pRefCon1, pRefCon2, knResultAborted);
        MessageBox("Out of memory");
        OnClicked_Close();
    }
    END_CATCH_ALL
}  // TMainDialog::ProcessPixelMessage


void TMainDialog::ProcessPixelMessage_Completion(void* pTMainDialog, void* pStruct, int nResult)
{
    TMainDialog*                 pThis                = (TMainDialog*)pTMainDialog;
    LPPROCESSPIXELMESSAGE_ASYNC  pProcessPixelMessage = (LPPROCESSPIXELMESSAGE_ASYNC)pStruct;

    ASSERT(pThis != NULL);
    ASSERT(pProcessPixelMessage != NULL);
    ASSERT(pProcessPixelMessage->param_pfnCompletion != NULL);

    if(nResult == knResultOK)
    {
	    CDC*  pDC = pThis->GetDC();
	    if(pDC != NULL)
	    {
	        // draw single pixel
	        pDC->SetPixel(pProcessPixelMessage->local_toRead.x,
	                      pProcessPixelMessage->local_toRead.y,
	                      pProcessPixelMessage->local_toRead.color);
	        pDC->SetPixel(pProcessPixelMessage->local_toRead.x,
	                      pProcessPixelMessage->local_toRead.y + 1,
	                      pProcessPixelMessage->local_toRead.color);
	        pDC->SetPixel(pProcessPixelMessage->local_toRead.x + 1,
	                      pProcessPixelMessage->local_toRead.y,
	                      pProcessPixelMessage->local_toRead.color);
	        pDC->SetPixel(pProcessPixelMessage->local_toRead.x + 1,
	                      pProcessPixelMessage->local_toRead.y + 1,
	                      pProcessPixelMessage->local_toRead.color);
	
	        pThis->ReleaseDC(pDC);
	    }
	}

    (*pProcessPixelMessage->param_pfnCompletion)(pProcessPixelMessage->param_pRefCon1,
                                                 pProcessPixelMessage->param_pRefCon2,
                                                 nResult);

    delete pProcessPixelMessage;
    pProcessPixelMessage = NULL;
}  // TMainDialog::ProcessPixelMessage_Completion


typedef struct tagPROCESSLINEMESSAGE_ASYNC
{
    CompletionFn  param_pfnCompletion;
    void*         param_pRefCon1;
    void*         param_pRefCon2;
    LINE_DATA     local_toRead;
} PROCESSLINEMESSAGE_ASYNC, * LPPROCESSLINEMESSAGE_ASYNC;


void TMainDialog::ProcessLineMessage(CompletionFn pfnCompletion, void* pRefCon1, void* pRefCon2)
{
    ASSERT(pfnCompletion != NULL);

    LPPROCESSLINEMESSAGE_ASYNC  pProcessLineMessage;
    
    TRY
    {
        pProcessLineMessage = new PROCESSLINEMESSAGE_ASYNC;
        pProcessLineMessage->param_pfnCompletion = pfnCompletion;
        pProcessLineMessage->param_pRefCon1      = pRefCon1;
        pProcessLineMessage->param_pRefCon2      = pRefCon2;

        m_CommunicationSocket.Read(&pProcessLineMessage->local_toRead,
                                   sizeof(pProcessLineMessage->local_toRead),
                                   ProcessLineMessage_Completion,
                                   this,
                                   pProcessLineMessage);
    }
    CATCH_ALL(e)
    {
        // let the callback do its cleanup
        (*pfnCompletion)(pRefCon1, pRefCon2, knResultAborted);
        MessageBox("Out of memory");
        OnClicked_Close();
    }
    END_CATCH_ALL
}  // TMainDialog::ProcessLineMessage


void TMainDialog::ProcessLineMessage_Completion(void* pTMainDialog, void* pStruct, int nResult)
{
    TMainDialog*                pThis               = (TMainDialog*)pTMainDialog;
    LPPROCESSLINEMESSAGE_ASYNC  pProcessLineMessage = (LPPROCESSLINEMESSAGE_ASYNC)pStruct;

    ASSERT(pThis != NULL);
    ASSERT(pProcessLineMessage != NULL);
    ASSERT(pProcessLineMessage->param_pfnCompletion != NULL);

    if(nResult == knResultOK)
    {
	    CDC*  pDC = pThis->GetDC();
	    if(pDC != NULL)
	    {
	        if(pDC->SaveDC() != 0)
	        {
	            CPen  ColorPen(PS_SOLID,
	                           pProcessLineMessage->local_toRead.nPenWidth,
	                           pProcessLineMessage->local_toRead.color);
	    
	            // draw line
	            pDC->SelectObject(&ColorPen);
	            pDC->SetROP2(R2_COPYPEN);
	            pDC->MoveTo(pProcessLineMessage->local_toRead.x1,
	                        pProcessLineMessage->local_toRead.y1);
	            pDC->LineTo(pProcessLineMessage->local_toRead.x2,
	                        pProcessLineMessage->local_toRead.y2);
	            
	            pDC->RestoreDC(-1);
	        }
	        pThis->ReleaseDC(pDC);
	    }
	}

    (*pProcessLineMessage->param_pfnCompletion)(pProcessLineMessage->param_pRefCon1,
                                                pProcessLineMessage->param_pRefCon2,
                                                nResult);

    delete pProcessLineMessage;
    pProcessLineMessage = NULL;
}  // TMainDialog::ProcessLineMessage_Completion


typedef struct tagTRANSMITDONEMESSAGE_ASYNC
{
    CompletionFn  param_pfnCompletion;
    void*         param_pRefCon1;
    void*         param_pRefCon2;
    HEADER_DATA   local_toTransmit;
} TRANSMITDONEMESSAGE_ASYNC, * LPTRANSMITDONEMESSAGE_ASYNC;


BOOL TMainDialog::TransmitDoneMessage(CompletionFn pfnCompletion, void* pRefCon1, void* pRefCon2)
{
    ASSERT(pfnCompletion != NULL);
    ASSERT(m_bConnected);
    ASSERT(m_bDrawingSide);

    LPTRANSMITDONEMESSAGE_ASYNC  pTransmitDoneMessage;
    
    if(m_CommunicationSocket.IsWriteOutstanding())
    {
        return(FALSE);
    }
    
    TRY
    {
        pTransmitDoneMessage = new TRANSMITDONEMESSAGE_ASYNC;
        pTransmitDoneMessage->param_pfnCompletion = pfnCompletion;
        pTransmitDoneMessage->param_pRefCon1      = pRefCon1;
        pTransmitDoneMessage->param_pRefCon2      = pRefCon2;
        pTransmitDoneMessage->local_toTransmit.message = kchDoneMessage;

        m_CommunicationSocket.Write(&pTransmitDoneMessage->local_toTransmit,
                                    sizeof(pTransmitDoneMessage->local_toTransmit),
                                    TransmitDoneMessage_Completion,
                                    this,
                                    pTransmitDoneMessage);
    }
    CATCH_ALL(e)
    {
        MessageBox("Out of memory");
        return(FALSE);
    }
    END_CATCH_ALL
    
    return(TRUE);
}  // TMainDialog::TransmitDoneMessage


void TMainDialog::TransmitDoneMessage_Completion(void* pTMainDialog, void* pStruct, int nResult)
{
    TMainDialog*                 pThis                = (TMainDialog*)pTMainDialog;
    LPTRANSMITDONEMESSAGE_ASYNC  pTransmitDoneMessage = (LPTRANSMITDONEMESSAGE_ASYNC)pStruct;

    ASSERT(pThis != NULL);
    ASSERT(pTransmitDoneMessage != NULL);
    ASSERT(pTransmitDoneMessage->param_pfnCompletion != NULL);

    (*pTransmitDoneMessage->param_pfnCompletion)(pTransmitDoneMessage->param_pRefCon1,
                                                 pTransmitDoneMessage->param_pRefCon2,
                                                 nResult);

    delete pTransmitDoneMessage;
    pTransmitDoneMessage = NULL;
}  // TMainDialog::TransmitDoneMessage_Completion


typedef struct tagTRANSMITPIXELMESSAGE_ASYNC
{
    CompletionFn  param_pfnCompletion;
    void*         param_pRefCon1;
    void*         param_pRefCon2;
    PIXEL_DATA    param_pixelData;  // copy of entire structure
    HEADER_DATA   local_toTransmit;
} TRANSMITPIXELMESSAGE_ASYNC, * LPTRANSMITPIXELMESSAGE_ASYNC;


BOOL TMainDialog::TransmitPixelMessage(LPPIXEL_DATA lpData, CompletionFn pfnCompletion, void* pRefCon1, void* pRefCon2)
{
    ASSERT(lpData != NULL);
    ASSERT(pfnCompletion != NULL);
    ASSERT(m_bConnected);
    ASSERT(m_bDrawingSide);
    
    LPTRANSMITPIXELMESSAGE_ASYNC  pTransmitPixelMessage;
    
    if(m_CommunicationSocket.IsWriteOutstanding())
    {
        return(FALSE);
    }
    
    TRY
    {
        pTransmitPixelMessage = new TRANSMITPIXELMESSAGE_ASYNC;
        pTransmitPixelMessage->param_pfnCompletion = pfnCompletion;
        pTransmitPixelMessage->param_pRefCon1      = pRefCon1;
        pTransmitPixelMessage->param_pRefCon2      = pRefCon2;
        pTransmitPixelMessage->param_pixelData     = *lpData;  // copy structure
        pTransmitPixelMessage->local_toTransmit.message = kchPixelMessage;

        m_CommunicationSocket.Write(&pTransmitPixelMessage->local_toTransmit,
                                    sizeof(pTransmitPixelMessage->local_toTransmit),
                                    TransmitPixelMessage_Completion1,
                                    this,
                                    pTransmitPixelMessage);
    }
    CATCH_ALL(e)
    {
        MessageBox("Out of memory");
        return(FALSE);
    }
    END_CATCH_ALL
    
    return(TRUE);
}  // TMainDialog::TransmitPixelMessage


void TMainDialog::TransmitPixelMessage_Completion1(void* pTMainDialog, void* pStruct, int nResult)
{
    TMainDialog*                  pThis                 = (TMainDialog*)pTMainDialog;
    LPTRANSMITPIXELMESSAGE_ASYNC  pTransmitPixelMessage = (LPTRANSMITPIXELMESSAGE_ASYNC)pStruct;

    ASSERT(pThis != NULL);
    ASSERT(pTransmitPixelMessage != NULL);
    ASSERT(pTransmitPixelMessage->param_pfnCompletion != NULL);

    if(nResult == knResultOK)
    {
	    pThis->m_CommunicationSocket.Write(&pTransmitPixelMessage->param_pixelData,
	                                       sizeof(pTransmitPixelMessage->param_pixelData),
	                                       TransmitPixelMessage_Completion2,
	                                       pThis,
	                                       pTransmitPixelMessage);
    }
    else
    {
	    (*pTransmitPixelMessage->param_pfnCompletion)(pTransmitPixelMessage->param_pRefCon1,
	                                                  pTransmitPixelMessage->param_pRefCon2,
	                                                  nResult);
	
	    delete pTransmitPixelMessage;
	    pTransmitPixelMessage = NULL;
        return;
    }
}  // TMainDialog::TransmitPixelMessage_Completion1


void TMainDialog::TransmitPixelMessage_Completion2(void* pTMainDialog, void* pStruct, int nResult)
{
    TMainDialog*                  pThis                 = (TMainDialog*)pTMainDialog;
    LPTRANSMITPIXELMESSAGE_ASYNC  pTransmitPixelMessage = (LPTRANSMITPIXELMESSAGE_ASYNC)pStruct;

    ASSERT(pThis != NULL);
    ASSERT(pTransmitPixelMessage != NULL);
    ASSERT(pTransmitPixelMessage->param_pfnCompletion != NULL);

    (*pTransmitPixelMessage->param_pfnCompletion)(pTransmitPixelMessage->param_pRefCon1,
                                                  pTransmitPixelMessage->param_pRefCon2,
                                                  nResult);

    delete pTransmitPixelMessage;
    pTransmitPixelMessage = NULL;
}  // TMainDialog::TransmitPixelMessage_Completion2


typedef struct tagTRANSMITLINEMESSAGE_ASYNC
{
    CompletionFn  param_pfnCompletion;
    void*         param_pRefCon1;
    void*         param_pRefCon2;
    LINE_DATA     param_lineData;  // copy of entire structure
    HEADER_DATA   local_toTransmit;
} TRANSMITLINEMESSAGE_ASYNC, * LPTRANSMITLINEMESSAGE_ASYNC;


BOOL TMainDialog::TransmitLineMessage(LPLINE_DATA lpData, CompletionFn pfnCompletion, void* pRefCon1, void* pRefCon2)
{
    ASSERT(lpData != NULL);
    ASSERT(pfnCompletion != NULL);
    ASSERT(m_bConnected);
    ASSERT(m_bDrawingSide);
    
    LPTRANSMITLINEMESSAGE_ASYNC  pTransmitLineMessage;
    
    if(m_CommunicationSocket.IsWriteOutstanding())
    {
        return(FALSE);
    }
    
    TRY
    {
        pTransmitLineMessage = new TRANSMITLINEMESSAGE_ASYNC;
        pTransmitLineMessage->param_pfnCompletion = pfnCompletion;
        pTransmitLineMessage->param_pRefCon1      = pRefCon1;
        pTransmitLineMessage->param_pRefCon2      = pRefCon2;
        pTransmitLineMessage->param_lineData      = *lpData;  // copy structure
        pTransmitLineMessage->local_toTransmit.message = kchLineMessage;

        m_CommunicationSocket.Write(&pTransmitLineMessage->local_toTransmit,
                                    sizeof(pTransmitLineMessage->local_toTransmit),
                                    TransmitLineMessage_Completion1,
                                    this,
                                    pTransmitLineMessage);
    }
    CATCH_ALL(e)
    {
        MessageBox("Out of memory");
        return(FALSE);
    }
    END_CATCH_ALL
    
    return(TRUE);
}  // TMainDialog::TransmitLineMessage


void TMainDialog::TransmitLineMessage_Completion1(void* pTMainDialog, void* pStruct, int nResult)
{
    TMainDialog*                 pThis                = (TMainDialog*)pTMainDialog;
    LPTRANSMITLINEMESSAGE_ASYNC  pTransmitLineMessage = (LPTRANSMITLINEMESSAGE_ASYNC)pStruct;

    ASSERT(pThis != NULL);
    ASSERT(pTransmitLineMessage != NULL);
    ASSERT(pTransmitLineMessage->param_pfnCompletion != NULL);

    if(nResult == knResultOK)
    {
	    pThis->m_CommunicationSocket.Write(&pTransmitLineMessage->param_lineData,
	                                       sizeof(pTransmitLineMessage->param_lineData),
	                                       TransmitLineMessage_Completion2,
	                                       pThis,
	                                       pTransmitLineMessage);
    }
    else
    {
	    (*pTransmitLineMessage->param_pfnCompletion)(pTransmitLineMessage->param_pRefCon1,
	                                                 pTransmitLineMessage->param_pRefCon2,
	                                                 nResult);
	
	    delete pTransmitLineMessage;
	    pTransmitLineMessage = NULL;
        return;
    }
}  // TMainDialog::TransmitLineMessage_Completion1


void TMainDialog::TransmitLineMessage_Completion2(void* pTMainDialog, void* pStruct, int nResult)
{
    TMainDialog*                 pThis                 = (TMainDialog*)pTMainDialog;
    LPTRANSMITLINEMESSAGE_ASYNC  pTransmitLineMessage = (LPTRANSMITLINEMESSAGE_ASYNC)pStruct;

    ASSERT(pThis != NULL);
    ASSERT(pTransmitLineMessage != NULL);
    ASSERT(pTransmitLineMessage->param_pfnCompletion != NULL);

    (*pTransmitLineMessage->param_pfnCompletion)(pTransmitLineMessage->param_pRefCon1,
                                                 pTransmitLineMessage->param_pRefCon2,
                                                 nResult);

    delete pTransmitLineMessage;
    pTransmitLineMessage = NULL;
}  // TMainDialog::TransmitLineMessage_Completion2


void TMainDialog::DoNothing_Completion(void* pTMainDialog, void* pUnused, int nResult)
{
    // do nothing    
}  // TMainDialog::DoNothing_Completion


void TMainDialog::OnClicked_Close_Completion(void* pTMainDialog, void* pUnused, int nResult)
{
    TMainDialog*  pThis = (TMainDialog*)pTMainDialog;
    
    ASSERT(pThis != NULL);

    pThis->m_CommunicationSocket.Close();
    pThis->m_bConnected = FALSE;
    pThis->UpdateControls();
}  // TMainDialog::OnClicked_Close_Completion


BOOL TMainDialog::OnInitDialog()
{
    CDialog::OnInitDialog();
    UpdateControls();
    return(TRUE);  // did not explicitly set focus
}  // TMainDialog::OnInitDialog


void TMainDialog::OnOK()
{
}  // TMainDialog::OnOK


void TMainDialog::OnCancel()
{
}  // TMainDialog::OnCancel


BEGIN_MESSAGE_MAP(TMainDialog, CDialog)
    ON_WM_CLOSE()
    ON_WM_LBUTTONDOWN()
    ON_WM_LBUTTONUP()
    ON_WM_MOUSEMOVE()
    ON_WM_RBUTTONDOWN()
    ON_BN_CLICKED(IDC_BUTTON_CONNECT, OnClicked_Connect)
    ON_BN_CLICKED(IDC_BUTTON_LISTEN,  OnClicked_Listen)
    ON_BN_CLICKED(IDC_BUTTON_CLOSE,   OnClicked_Close)
    ON_MESSAGE(WMU_SOCKET_CONNECT,    OnSocketConnect)
    ON_MESSAGE(WMU_SOCKET_CLOSED,     OnSocketClosed)
    ON_MESSAGE(WMU_SOCKET_ERROR,      OnSocketError)
END_MESSAGE_MAP()


afx_msg void TMainDialog::OnClose()
{
    if(m_bConnected)
    {
        OnClicked_Close();
    }
    
    EndDialog(IDOK);
}  // TMainDialog::OnClose


afx_msg void TMainDialog::OnLButtonDown(UINT nFlags, CPoint point)
{
    if(!m_bConnected || !m_bDrawingSide)
    {
        return;
    }
    
    if(m_CommunicationSocket.IsWriteOutstanding())
    {
        return;
    }
    
    ASSERT(!m_bCaptured);
    
    SetCapture();
    m_bCaptured = TRUE;
        
    m_bIsLine = FALSE;
    m_ptDown = point;
}  // TMainDialog::OnLButtonDown


afx_msg void TMainDialog::OnLButtonUp(UINT nFlags, CPoint point)
{
    if(!m_bConnected || !m_bDrawingSide || !m_bCaptured)
    {
        return;
    }

    CDC*  pDC = GetDC();
    if(pDC != NULL)
    {
        if(pDC->SaveDC() != 0)
        {
            if(m_bIsLine)
            {
                CPen  ColorPen(PS_SOLID, knPenWidth, m_DrawColor);
                
                // erase old line
                pDC->SelectObject(&ColorPen);
                pDC->SetROP2(R2_NOTXORPEN);
                pDC->MoveTo(m_ptDown.x, m_ptDown.y);
                pDC->LineTo(m_ptPrevious.x, m_ptPrevious.y);
                
                // draw real line
                pDC->SetROP2(R2_COPYPEN);
                pDC->MoveTo(m_ptDown.x, m_ptDown.y);
                pDC->LineTo(point.x, point.y);
            }
            else
            {
                // draw single pixel
                pDC->SetPixel(m_ptDown.x    , m_ptDown.y    , m_DrawColor);
                pDC->SetPixel(m_ptDown.x    , m_ptDown.y + 1, m_DrawColor);
                pDC->SetPixel(m_ptDown.x + 1, m_ptDown.y    , m_DrawColor);
                pDC->SetPixel(m_ptDown.x + 1, m_ptDown.y + 1, m_DrawColor);
            }
            pDC->RestoreDC(-1);
        }
        ReleaseDC(pDC);
    }

    ReleaseCapture();
    m_bCaptured = FALSE;

    if(m_bIsLine)
    {
        // transmit line data
        LINE_DATA  toTransmit;
        toTransmit.x1        = m_ptDown.x;
        toTransmit.y1        = m_ptDown.y;
        toTransmit.x2        = point.x;
        toTransmit.y2        = point.y;
        toTransmit.color     = m_DrawColor;
        toTransmit.nPenWidth = knPenWidth;

        if(!TransmitLineMessage(&toTransmit,
                                DoNothing_Completion,
                                this,
                                NULL))
        {
            // connection went down
            OnSocketError(0, 0);
        }
    }
    else
    {
        // transmit pixel data
        PIXEL_DATA  toTransmit;
        toTransmit.x     = m_ptDown.x;
        toTransmit.y     = m_ptDown.y;
        toTransmit.color = m_DrawColor;

        if(!TransmitPixelMessage(&toTransmit,
                                 DoNothing_Completion,
                                 this,
                                 NULL))
        {
            // connection went down
            OnSocketError(0, 0);
        }
    }
}  // TMainDialog::OnLButtonUp


afx_msg void TMainDialog::OnMouseMove(UINT nFlags, CPoint point)
{
    if(!m_bConnected || !m_bDrawingSide || !m_bCaptured)
    {
        return;
    }

    CDC*  pDC = GetDC();
    if(pDC != NULL)
    {
        if(pDC->SaveDC() != 0)
        {
            CPen  ColorPen(PS_SOLID, knPenWidth, m_DrawColor);

            if(m_bIsLine)
            {
                // erase old line
                pDC->SelectObject(&ColorPen);
                pDC->SetROP2(R2_NOTXORPEN);
                pDC->MoveTo(m_ptDown.x, m_ptDown.y);
                pDC->LineTo(m_ptPrevious.x, m_ptPrevious.y);
                
                // draw new line
                pDC->MoveTo(m_ptDown.x, m_ptDown.y);
                pDC->LineTo(point.x, point.y);
        
                m_ptPrevious = point;
            }
            else
            {
                if(abs(point.x - m_ptDown.x) > 2 ||
                   abs(point.y - m_ptDown.y) > 2)
                {
                    // draw first line
                    pDC->SelectObject(&ColorPen);
                    pDC->SetROP2(R2_NOTXORPEN);
                    pDC->MoveTo(m_ptDown.x, m_ptDown.y);
                    pDC->LineTo(point.x, point.y);
        
                    m_bIsLine = TRUE;
                    m_ptPrevious = point;
                }
            }
            pDC->RestoreDC(-1);
        }
        ReleaseDC(pDC);
    }
}  // TMainDialog::OnMouseMove


afx_msg void TMainDialog::OnRButtonDown(UINT nFlags, CPoint point)
{
    if(!m_bCaptured)
    {
        CColorDialog  dlgColor(m_DrawColor, 0, this);
            
        if(dlgColor.DoModal() == IDOK)
        {
            m_DrawColor = dlgColor.GetColor();
        }
    }
}  // TMainDialog::OnRButtonDown


afx_msg void TMainDialog::OnClicked_Connect()
{
    ASSERT(!m_bConnected);
    ASSERT(!m_CommunicationSocket.IsWriteOutstanding());
    
    TCHAR  szIPAddress[128];
    UINT   nPort;
    BOOL   bConnecting = FALSE;

    GetDlgItemText(IDC_EDIT_IPADDRESS, szIPAddress, sizeof(szIPAddress) / sizeof(szIPAddress[0]));
    nPort = GetDlgItemInt(IDC_EDIT_PORT, NULL, FALSE);

    if(m_CommunicationSocket.Create())
    {
        // expect Connect() to fail with WSAWOULDBLOCK
        
        if(!m_CommunicationSocket.Connect(szIPAddress, nPort))
        {
            if(m_CommunicationSocket.GetLastError() == WSAEWOULDBLOCK)
            {
                // will be notified in OnSocketConnect()
                GetDlgItem(IDC_BUTTON_CONNECT)->EnableWindow(FALSE);
                GetDlgItem(IDC_BUTTON_LISTEN)->EnableWindow(FALSE);
                GetDlgItem(IDC_BUTTON_CLOSE)->EnableWindow(FALSE);
                bConnecting = TRUE;
            }
            else
            {
                m_CommunicationSocket.Close();
            }
        }
        else
        {
            m_CommunicationSocket.Close();
        }
    }
    
    if(!bConnecting)
    {
        MessageBox("Unable to connect to remote machine.");
    }
}  // TMainDialog::OnClicked_Connect


afx_msg void TMainDialog::OnClicked_Listen()
{
    ASSERT(!m_bConnected);

    // do a sync listen and accept for simplicity...
    
    CSocket  ListenSocket;
    UINT     nPort;

    nPort = GetDlgItemInt(IDC_EDIT_PORT, NULL, FALSE);

    if(ListenSocket.Create(nPort))
    {
        if(ListenSocket.Listen() && ListenSocket.Accept(m_CommunicationSocket))
        {
            m_bConnected = TRUE;
            m_bDrawingSide = FALSE;
        }
    }
    
    if(m_bConnected)
    {
        UpdateControls();
        ReadSocketData();
    }
    else
    {
        MessageBox("Failed to establish connection with remote machine.");
    }
}  // TMainDialog::OnClicked_Listen


afx_msg void TMainDialog::OnClicked_Close()
{
    ASSERT(m_bConnected);
    
    BOOL  bTransmittedDone = FALSE;
    
    if(m_bDrawingSide)
    {
        // try to notify the other side that we're shutting down
        bTransmittedDone = TransmitDoneMessage(OnClicked_Close_Completion,
                                               this,
                                               NULL);
    }

    if(!bTransmittedDone)
    {
        OnClicked_Close_Completion(this, NULL, knResultOK);
    }
}  // TMainDialog::OnClicked_Close


afx_msg LONG TMainDialog::OnSocketConnect(WPARAM wParam, LPARAM lParam)
{
    ASSERT((TAsyncSocket*)lParam == &m_CommunicationSocket);

    if(wParam == 0)
    {
        // established connection
        m_bConnected = TRUE;
        m_bDrawingSide = TRUE;
        UpdateControls();
    }
    else
    {
        // no need to close the socket since it closed itself
        //   before posting this message
        UpdateControls();
        MessageBox("Unable to connect to remote machine.");
    }
    
    return(0);
}  // TMainDialog::OnSocketConnect


afx_msg LONG TMainDialog::OnSocketClosed(WPARAM wParam, LPARAM lParam)
{
    ASSERT((TAsyncSocket*)lParam == &m_CommunicationSocket);

    // no need to close the socket since it closed itself
    //   before posting this message
    m_bConnected = FALSE;
    UpdateControls();
    MessageBox("Remote machine closed down the socket");

    return(0);
}  // TMainDialog::OnSocketClosed


afx_msg LONG TMainDialog::OnSocketError(WPARAM wParam, LPARAM lParam)
{
    ASSERT((TAsyncSocket*)lParam == &m_CommunicationSocket);

    // no need to close the socket since it closed itself
    //   before posting this message
    m_bConnected = FALSE;
    UpdateControls();
    MessageBox("A socket error occurred");

    return(0);
}  // TMainDialog::OnSocketError
