#include "prechdr.hpp"
#include "maindlg.hpp"


static const int   knPenWidth              = 2;
static const UINT  WMU_SOCKET_NOTIFICATION = WM_USER + 101;


TMainDialog::TMainDialog(CWnd* pParent)
    : CDialog(IDD_MAINWINDOW, pParent)
{
    m_bConnected = FALSE;
    m_bCaptured  = FALSE;
    m_DrawColor  = RGB(0, 0, 0);  // black

    m_CommunicationSocket.SetNotification(this, WMU_SOCKET_NOTIFICATION);
}  // TMainDialog::TMainDialog


TMainDialog::~TMainDialog()
{
}  // TMainDialog::~TMainDialog


void TMainDialog::UpdateControls()
{
    GetDlgItem(IDC_BUTTON_CONNECT)->EnableWindow(!m_bConnected);
    GetDlgItem(IDC_BUTTON_LISTEN)->EnableWindow(!m_bConnected);
    GetDlgItem(IDC_BUTTON_CLOSE)->EnableWindow(m_bConnected);
}  // TMainDialog::UpdateControls


void TMainDialog::ReadSocketData()
{
    ASSERT(m_bConnected);

    HEADER_DATA  toRead;
    
    if(m_CommunicationSocket.Receive(&toRead, sizeof(toRead)) != sizeof(toRead))
    {
        OnClicked_Close();
        MessageBox("The connection was closed.");
        return;
    }

    while(toRead.message != kchDoneMessage)
    {
        if(!ProcessMessage(toRead.message))
        {
            OnClicked_Close();
	        MessageBox("The connection was closed.");
            return;
        }
        
        if(m_CommunicationSocket.Receive(&toRead, sizeof(toRead)) != sizeof(toRead))
        {
            OnClicked_Close();
	        MessageBox("The connection was closed.");
            return;
        }
    }
    
    // normal close
    OnClicked_Close();
    MessageBox("Finished watching other side.");
}  // TMainDialog::ReadSocketData


BOOL TMainDialog::ProcessMessage(char chMessage)
{
    BOOL  bConnectionOK = FALSE;
    
    switch(chMessage)
    {
        case(kchPixelMessage):
            bConnectionOK = ProcessPixelMessage();
            break;
        case(kchLineMessage):
            bConnectionOK = ProcessLineMessage();
            break;
    }
    
    return(bConnectionOK);
}  // TMainDialog::ProcessMessage


BOOL TMainDialog::ProcessPixelMessage()
{
    PIXEL_DATA  toRead;
    
    if(m_CommunicationSocket.Receive(&toRead, sizeof(toRead)) != sizeof(toRead))
    {
        return(FALSE);
    }
    
    CDC*  pDC = GetDC();
    if(pDC != NULL)
    {
        // draw single pixel
        pDC->SetPixel(toRead.x,     toRead.y    , toRead.color);
        pDC->SetPixel(toRead.x,     toRead.y + 1, toRead.color);
        pDC->SetPixel(toRead.x + 1, toRead.y    , toRead.color);
        pDC->SetPixel(toRead.x + 1, toRead.y + 1, toRead.color);
    
        ReleaseDC(pDC);
    }
    
    return(TRUE);
}  // TMainDialog::ProcessPixelMessage


BOOL TMainDialog::ProcessLineMessage()
{
    LINE_DATA  toRead;
    
    if(m_CommunicationSocket.Receive(&toRead, sizeof(toRead)) != sizeof(toRead))
    {
        return(FALSE);
    }

    CDC*  pDC = GetDC();
    if(pDC != NULL)
    {
        if(pDC->SaveDC() != 0)
        {
            CPen  ColorPen(PS_SOLID, toRead.nPenWidth, toRead.color);
    
            // draw line
            pDC->SelectObject(&ColorPen);
            pDC->SetROP2(R2_COPYPEN);
            pDC->MoveTo(toRead.x1, toRead.y1);
            pDC->LineTo(toRead.x2, toRead.y2);
            
            pDC->RestoreDC(-1);
        }
        ReleaseDC(pDC);
    }
    
    return(TRUE);
}  // TMainDialog::ProcessLineMessage


BOOL TMainDialog::TransmitDoneMessage()
{
    ASSERT(m_bConnected);
    ASSERT(m_bDrawingSide);
    
    HEADER_DATA  toTransmit;
    
    toTransmit.message = kchDoneMessage;
    if(m_CommunicationSocket.Send(&toTransmit, sizeof(toTransmit)) != sizeof(toTransmit))
    {
        return(FALSE);
    }

    return(TRUE);
}  // TMainDialog::TransmitDoneMessage


BOOL TMainDialog::TransmitPixelMessage(LPPIXEL_DATA lpData)
{
    ASSERT(lpData != NULL);
    ASSERT(m_bConnected);
    ASSERT(m_bDrawingSide);
    
    HEADER_DATA  toTransmit;
    
    toTransmit.message = kchPixelMessage;
    if(m_CommunicationSocket.Send(&toTransmit, sizeof(toTransmit)) != sizeof(toTransmit))
    {
        return(FALSE);
    }

    if(m_CommunicationSocket.Send(lpData, sizeof(PIXEL_DATA)) != sizeof(PIXEL_DATA))
    {
        return(FALSE);
    }
    
    return(TRUE);
}  // TMainDialog::TransmitPixelMessage


BOOL TMainDialog::TransmitLineMessage(LPLINE_DATA lpData)
{
    ASSERT(lpData != NULL);
    ASSERT(m_bConnected);
    ASSERT(m_bDrawingSide);

    HEADER_DATA  toTransmit;
    
    toTransmit.message = kchLineMessage;
    if(m_CommunicationSocket.Send(&toTransmit, sizeof(toTransmit)) != sizeof(toTransmit))
    {
        return(FALSE);
    }

    if(m_CommunicationSocket.Send(lpData, sizeof(LINE_DATA)) != sizeof(LINE_DATA))
    {
        return(FALSE);
    }
    
    return(TRUE);
}  // TMainDialog::TransmitLineMessage


BOOL TMainDialog::OnInitDialog()
{
    CDialog::OnInitDialog();
    UpdateControls();
    return(TRUE);  // did not explicitly set focus
}  // TMainDialog::OnInitDialog


void TMainDialog::OnOK()
{
}  // TMainDialog::OnOK


void TMainDialog::OnCancel()
{
}  // TMainDialog::OnCancel


BEGIN_MESSAGE_MAP(TMainDialog, CDialog)
    ON_WM_CLOSE()
    ON_WM_LBUTTONDOWN()
    ON_WM_LBUTTONUP()
    ON_WM_MOUSEMOVE()
    ON_WM_RBUTTONDOWN()
    ON_BN_CLICKED(IDC_BUTTON_CONNECT, OnClicked_Connect)
    ON_BN_CLICKED(IDC_BUTTON_LISTEN,  OnClicked_Listen)
    ON_BN_CLICKED(IDC_BUTTON_CLOSE,   OnClicked_Close)
    ON_MESSAGE(WMU_SOCKET_NOTIFICATION, OnSocketNotification)
END_MESSAGE_MAP()


afx_msg void TMainDialog::OnClose()
{
    EndDialog(IDOK);
}  // TMainDialog::OnClose


afx_msg void TMainDialog::OnLButtonDown(UINT nFlags, CPoint point)
{
    if(!m_bConnected || !m_bDrawingSide)
    {
        return;
    }
    
    ASSERT(!m_bCaptured);
    
    SetCapture();
    m_bCaptured = TRUE;
        
    m_bLine  = FALSE;
    m_ptDown = point;
}  // TMainDialog::OnLButtonDown


afx_msg void TMainDialog::OnLButtonUp(UINT nFlags, CPoint point)
{
    if(!m_bConnected || !m_bDrawingSide || !m_bCaptured)
    {
        return;
    }

    CDC*  pDC = GetDC();
    if(pDC != NULL)
    {
        if(pDC->SaveDC() != 0)
        {
            if(m_bLine)
            {
                CPen  ColorPen(PS_SOLID, knPenWidth, m_DrawColor);
                
                // erase old line
                pDC->SelectObject(&ColorPen);
                pDC->SetROP2(R2_NOTXORPEN);
                pDC->MoveTo(m_ptDown.x, m_ptDown.y);
                pDC->LineTo(m_ptPrevious.x, m_ptPrevious.y);
                
                // draw real line
                pDC->SetROP2(R2_COPYPEN);
                pDC->MoveTo(m_ptDown.x, m_ptDown.y);
                pDC->LineTo(point.x, point.y);
                
                // transmit line data
                LINE_DATA  toTransmit;
                toTransmit.x1        = m_ptDown.x;
                toTransmit.y1        = m_ptDown.y;
                toTransmit.x2        = point.x;
                toTransmit.y2        = point.y;
                toTransmit.color     = m_DrawColor;
                toTransmit.nPenWidth = knPenWidth;

                if(!TransmitLineMessage(&toTransmit))
                {
                    // @@@ connection went down
                    OnClicked_Close();
                }
            }
            else
            {
                // draw single pixel
                pDC->SetPixel(m_ptDown.x    , m_ptDown.y    , m_DrawColor);
                pDC->SetPixel(m_ptDown.x    , m_ptDown.y + 1, m_DrawColor);
                pDC->SetPixel(m_ptDown.x + 1, m_ptDown.y    , m_DrawColor);
                pDC->SetPixel(m_ptDown.x + 1, m_ptDown.y + 1, m_DrawColor);
        
                // transmit pixel data
                PIXEL_DATA  toTransmit;
                toTransmit.x     = m_ptDown.x;
                toTransmit.y     = m_ptDown.y;
                toTransmit.color = m_DrawColor;

                if(!TransmitPixelMessage(&toTransmit))
                {
                    // @@@ connection went down
                    OnClicked_Close();
                }
            }
            pDC->RestoreDC(-1);
        }
        ReleaseDC(pDC);
    }

    ReleaseCapture();
    m_bCaptured = FALSE;
}  // TMainDialog::OnLButtonUp


afx_msg void TMainDialog::OnMouseMove(UINT nFlags, CPoint point)
{
    if(!m_bConnected || !m_bDrawingSide || !m_bCaptured)
    {
        return;
    }

    CDC*  pDC = GetDC();
    if(pDC != NULL)
    {
        if(pDC->SaveDC() != 0)
        {
            CPen  ColorPen(PS_SOLID, knPenWidth, m_DrawColor);

            if(m_bLine)
            {
                // erase old line
                pDC->SelectObject(&ColorPen);
                pDC->SetROP2(R2_NOTXORPEN);
                pDC->MoveTo(m_ptDown.x, m_ptDown.y);
                pDC->LineTo(m_ptPrevious.x, m_ptPrevious.y);
                
                // draw new line
                pDC->MoveTo(m_ptDown.x, m_ptDown.y);
                pDC->LineTo(point.x, point.y);
        
                m_ptPrevious = point;
            }
            else
            {
                if(abs(point.x - m_ptDown.x) > 2 ||
                   abs(point.y - m_ptDown.y) > 2)
                {
                    // draw first line
                    pDC->SelectObject(&ColorPen);
                    pDC->SetROP2(R2_NOTXORPEN);
                    pDC->MoveTo(m_ptDown.x, m_ptDown.y);
                    pDC->LineTo(point.x, point.y);
        
                    m_bLine = TRUE;
                    m_ptPrevious = point;
                }
            }
            pDC->RestoreDC(-1);
        }
        ReleaseDC(pDC);
    }
}  // TMainDialog::OnMouseMove


afx_msg void TMainDialog::OnRButtonDown(UINT nFlags, CPoint point)
{
    if(!m_bCaptured)
    {
        CColorDialog  dlgColor(m_DrawColor, 0, this);
            
        if(dlgColor.DoModal() == IDOK)
        {
            m_DrawColor = dlgColor.GetColor();
        }
    }
}  // TMainDialog::OnRButtonDown


afx_msg void TMainDialog::OnClicked_Connect()
{
    ASSERT(!m_bConnected);
    
    TCHAR  szIPAddress[128];
    UINT   nPort;

    GetDlgItemText(IDC_EDIT_IPADDRESS, szIPAddress, sizeof(szIPAddress) / sizeof(szIPAddress[0]));
    nPort = GetDlgItemInt(IDC_EDIT_PORT, NULL, FALSE);

    if(m_CommunicationSocket.Create())
    {
        if(m_CommunicationSocket.Connect(szIPAddress, nPort))
        {
            m_bConnected = TRUE;
            m_bDrawingSide = TRUE;
        }
        else
        {
            m_CommunicationSocket.Close();
        }
    }
    
    if(m_bConnected)
    {
        UpdateControls();
    }
    else
    {
        MessageBox("Unable to connect to remote machine.");
    }
}  // TMainDialog::OnClicked_Connect


afx_msg void TMainDialog::OnClicked_Listen()
{
    ASSERT(!m_bConnected);

    CSocket  ListenSocket;
    UINT     nPort;

    nPort = GetDlgItemInt(IDC_EDIT_PORT, NULL, FALSE);

    if(ListenSocket.Create(nPort))
    {
        if(ListenSocket.Listen() && ListenSocket.Accept(m_CommunicationSocket))
        {
            m_bConnected = TRUE;
            m_bDrawingSide = FALSE;
        }
    }

    if(m_bConnected)
    {
        UpdateControls();
        ReadSocketData();
    }
    else
    {
        MessageBox("Failed to establish connection with remote machine.");
    }
}  // TMainDialog::OnClicked_Listen


afx_msg void TMainDialog::OnClicked_Close()
{
    ASSERT(m_bConnected);
    
    if(m_bConnected && m_bDrawingSide)
    {
        TransmitDoneMessage();
    }

    m_CommunicationSocket.Close();
    m_bConnected = FALSE;
    UpdateControls();
}  // TMainDialog::OnClicked_Close


afx_msg LONG TMainDialog::OnSocketNotification(WPARAM wParam, LPARAM lParam)
{
    ASSERT((TSyncSocket*)lParam == &m_CommunicationSocket);

    // no need to close the socket since it closed itself
    //   before posting this message
    m_bConnected = FALSE;
    UpdateControls();

    return(0);
}  // TMainDialog::OnSocketNotification
