#ifndef __TASOCKET_HPP__
#define __TASOCKET_HPP__


static const int  knResultOK      = 1;
static const int  knResultAborted = 2;


class TAsyncSocket;  // forward declaration
typedef void (* CompletionFn)(void* pRefCon1, void* pRefCon2, int nResult);


class TAsyncSocket : public CAsyncSocket
{
    public:
        TAsyncSocket();
        virtual ~TAsyncSocket();

        void  SetNotification(CWnd* pWindowToNotify, UINT unConnectMessage, UINT unClosedMessage, UINT unErrorMessage);

        BOOL  IsReadOutstanding() const;
        BOOL  IsWriteOutstanding() const;
        
        BOOL  Create(UINT nSocketPort = 0);
        BOOL  Initialize();
        void  Close();
        BOOL  Read(void* lpBuf, int nBufLen, CompletionFn pfnCompletion, void* pRefCon1, void* pRefCon2);
        BOOL  Write(const void* lpBuf, int nBufLen, CompletionFn pfnCompletion, void* pRefCon1, void* pRefCon2);

    protected:
        void  ProcessIncomingData();
        void  ProcessOutgoingData();

    protected:
        virtual void  OnConnect(int nErrorCode);
        virtual void  OnReceive(int nErrorCode);
        virtual void  OnSend(int nErrorCode);
        virtual void  OnClose(int nErrorCode);

    #ifdef _DEBUG
    protected:
        virtual void  AssertValid() const;
    #endif  // _DEBUG
        
    protected:
        // buffer of incoming data
        int           m_nIncomingBufferSize;
        char*         m_pIncomingBuffer;

        // outstanding read
        void*         m_pReadBuffer;
        int           m_nReadBufferLength;
        CompletionFn  m_pReadCompletionFn;
        void*         m_pReadRefCon1;
        void*         m_pReadRefCon2;

        // outstanding write
        const void*   m_pWriteBuffer;
        int           m_nWriteBufferLength;
        CompletionFn  m_pWriteCompletionFn;
        void*         m_pWriteRefCon1;
        void*         m_pWriteRefCon2;
        
        // socket owner information
        CWnd*         m_pWindowToNotify;
        UINT          m_unConnectMessage;
        UINT          m_unClosedMessage;
        UINT          m_unErrorMessage;

    private:
        // disable use of copy constructor or operator=
        TAsyncSocket(const TAsyncSocket&);
        void operator=(const TAsyncSocket&);
};  // TAsyncSocket


#endif  // __TASOCKET_HPP__
