#include "prechdr.hpp"
#include "tssocket.hpp"


TSyncSocket::TSyncSocket()
{
    m_pWindowToNotify       = NULL;
    m_unNotificationMessage = 0;
}  // TSyncSocket::TSyncSocket


TSyncSocket::~TSyncSocket()
{
}  // TSyncSocket::~TSyncSocket


void TSyncSocket::SetNotification(CWnd* pWindowToNotify, UINT unNotificationMessage)
{
    ASSERT(pWindowToNotify != NULL);
    ASSERT(unNotificationMessage >= WM_USER);
    
    m_pWindowToNotify       = pWindowToNotify;
    m_unNotificationMessage = unNotificationMessage;
    
    ASSERT_VALID(this);
}  // TSyncSocket::SetNotification


int TSyncSocket::Receive(void FAR* lpBuf, int nBufLen, int nFlags)
{
    ASSERT(lpBuf != NULL);
    ASSERT(nBufLen > 0);
    
    int  nLeftToRead = nBufLen;
    while(nLeftToRead > 0)
    {
        int  nRead = CSocket::Receive(lpBuf, nLeftToRead, nFlags);
        if(nRead == SOCKET_ERROR)
        {
            return(SOCKET_ERROR);
        }
        
        lpBuf = (char FAR*)lpBuf + nRead;
        nLeftToRead -= nRead;            
    }
    
    return(nBufLen);
}  // TSyncSocket::Receive


void TSyncSocket::OnClose(int nErrorCode)
{
    ASSERT_VALID(this);

    // other side closed socket, close and then tell the owner
    Close();
    m_pWindowToNotify->SendMessage(m_unNotificationMessage,
                                   (WPARAM)WSAECONNABORTED,
                                   MAKELPARAM((WORD)this, 0));
}  // TSyncSocket::OnClose


#ifdef _DEBUG
void TSyncSocket::AssertValid() const
{
    CSocket::AssertValid();
    
    ASSERT(m_pWindowToNotify != NULL);
    ASSERT(m_unNotificationMessage >= WM_USER);
}  // TSyncSocket::AssertValid
#endif  // _DEBUG
